'''
This is the Vivisect Extension component of VivisectION.

a Symnlink to this directory should be placed somewhere in your Vivisect Extensions path,
defined by the environment variable VIV_EXT_PATH and following typical pathing standards

eg.
$ export VIV_EXT_PATH=~/hacking/viv_extensions:~/work/viv_ext
$ ln -s <path_to_this_dir> ~/hacking/viv_extensions/
'''
import os
import sys
import uuid
import threading
import traceback

import envi.interactive as ei
import envi.config as e_config
import envi.threads as e_thread

import vivisection.viv_plugin.share as ion_share


from PyQt5.QtWidgets import QToolBar, QLabel, QPushButton, QTextEdit, QWidget, QInputDialog
from PyQt5 import QtCore

from vqt.main import idlethread
from vqt.basics import VBox
from vqt.common import ACT



### standard UI stuff
def vprint(vw, s, *args, **kwargs):
    vw.vprint(s % args)
    print(s % args)

def ctxMenuHook(vw, va, expr, menu, parent, nav):
    '''
    Context Menu handler (adds options as we wish)
    '''
    try:
        if va == vw.getFunction(va):
            # FIXME: make shared workspace mode work correctly, then uncomment:
            #menu.addAction('SmartEmu', ACT(launchEmuShared, vw, "0x%x"%va))
            menu.addAction('SmartEmu - console', ACT(launchEmuLocal, vw, "0x%x"%va))

    except Exception as e:
        traceback.print_exc()

class IonToolbar(QToolBar):
    def __init__(self, vw, vwgui):
        self.vw = vw
        self.vwgui = vwgui

        QToolBar.__init__(self, parent=vwgui)
        self.addWidget( QLabel('Ion tools:', parent=self) )
        self.addAction('CLI', self.cli)

    def cli(self):
        # TODO: make this and FuncEmulator 1-shot/mutex, since there can be only one (console)
        vw = self.vw
        vwgui = self.vwgui

        self.vw.clithread = threading.Thread(target=ei.dbg_interact, args=(locals(), globals()), daemon=True)
        self.vw.clithread.start()

### emulation support tools
DEFAULT_SETUPCODE = '''fva = vw.parseExpression('%s')
import vivisection.emutils as vemu
emu = vw.getEmulator(va=fva, logread=True, logwrite=True, safemem=False)
nemu = vemu.NinjaEmulator(emu, fakePEB=True, guiFuncGraphName="FuncGraph0")

args = []
nemu.setupCall(fva, args)
nemu.runStep()
'''

defconfig = {
        'termcmd': 'konsole -e <vwcmd>',
        'setup_code': DEFAULT_SETUPCODE,
        'cache': {},

        }
class IonManager:
    def __init__(self, vw, autosave=True):
        self.vw = vw
        self.sessions = {}
        self.console_in_use = False

        self.ionhome = e_config.gethomedir('.ion')

        # Load up the config
        cfgfile = os.path.join(self.ionhome, 'vivisection.json')
        self.config = e_config.EnviConfig(filename=cfgfile, defaults=defconfig, autosave=autosave)

        # share out the workspace
        # FIXME: check if connected to a VivServer and support that instead
        # FIXME: setting this as the server causes saving to stop working.  need to work around another method of enabling the "Follow/Lead" in the gui.  probably an upstream problem to solve.  for now, we have the console.
        #self.daemon = ion_share.shareWorkspace(vw)

    def addSession(self, sessid, sesstup):
        if sessid in self.sessions:
            logger.warning("IonManager:addSession refusing to add an existing session (%r).", sessid)
            return

        self.sessions[sessid] = sesstup

    def getSession(self, sessid):
        if sessid not in self.sessions:
            logger.warning("IonManager:getSession %r session does not exist.", sessid)
            return

        return self.sessions[sessid]

    def isConsoleInUse(self):
        return self.console_in_use

    def setUsingConsole(self, state=True):
        self.console_in_use = state

def getSetupCode(vw, fvaexpr):
    '''
    Caches results in Ion config
    '''
    defcode = vw._ionmgr.config.setup_code
    defcode = defcode % fvaexpr
    fva = vw.parseExpression(fvaexpr)
    initialcode = vw._ionmgr.config.cache.get(fva, defcode)

    setup_code, ok = QInputDialog.getMultiLineText(None, 'Enter Setup Code', 'Code:', text=initialcode)
    if ok:
        vw._ionmgr.config.cache.fva = setup_code
        vw._ionmgr.config.saveConfigFile()
        return setup_code
  
def launchEmuShared(vw, fvaexpr):
    # TODO: fix up Vivisect's Shared Workspace.  there are a few bugs remaining.
    #### THIS DOESN'T WORK CORRECTLY YET....
    sessid = uuid.uuid1().hex
    ctx = {}
    setup_code = getSetupCode(vw, fvaexpr)
    if setup_code is None:
        return

    ctx['setup_code'] = setup_code
    
    vw._ionmgr.addSession(sessid, ctx)
    port = vw._ionmgr.daemon.port

    termcmd = vw._ionmgr.config.termcmd #'konsole -e <vwcmd>'

    cmdsyntax = '%s -s %s %s %s' % (sys.argv[0], 'localhost', port, sessid)
    termcmd = termcmd.replace('<vwcmd>', cmdsyntax)
    os.system(termcmd)

def launchEmuLocal(vw, fvaexpr):
    # setup and popup
    setup_code = getSetupCode(vw, fvaexpr)
    if setup_code is None:
        return

    # FIXME: track and check if any other thing is using the console and pop up 
    # an error message if it is busy.

    # execute setup
    if vw._ionmgr.isConsoleInUse():
        # pop up message
        print("\n\n=== Can't use console, already in use!")
        return

    vw._ionmgr.setUsingConsole(True)
    t = threading.Thread(target=runInteractive, \
            args=(vw, fvaexpr, setup_code, globals(), locals()), daemon=True)
    t.start()

    # we can also do the external gui widget, but not until we can replace the gui mainloop with ipython

def runInteractive(vw, fvaexpr, setup_code, gbls, lcls):
    '''
    Do the interactive emulation and end with an IPython shell.
    This function is specifically to group the emulation and ipython so they
    can be run in a separate thread.
    '''
    exec(setup_code, gbls, lcls)
    ei.dbg_interact(lcls, gbls, "VivisectION Emulation.  nemu.runStep() to start.")

    # very important: Free up Console for reuse
    vw._ionmgr.setUsingConsole(False)


@idlethread
def vivExtension(vw, vwgui):
    # Add IonManager into the workspace:
    vw._ionmgr = IonManager(vw)

    # Create a toolbar and add it to the GUI
    toolbar = IonToolbar(vw, vwgui)
    vwgui.addToolBar(QtCore.Qt.TopToolBarArea, toolbar)

    # Add a menu item
    vwgui.vqAddMenuField('&Plugins.&Ion.&PrintDiscoveredStats', vw.printDiscoveredStats, ())

    # hook context menu
    vw.addCtxMenuHook('ion', ctxMenuHook)





def getPosixNinjaEmu(emu, fva=None, funcgraph='FuncGraph0'):
    '''
    Simple helper script to simplify and bring consistency to all functions
    '''
    vemu.getHeap(emu, 5024*1024)    # initialize heap to 1MB instead of 10k default
    nemu = vemu.NinjaEmulator(emu, fakePEB=True, guiFuncGraphName=funcgraph)

def getWinBaseEmu(emu, fva=None, funcgraph='FuncGraph0'):
    '''
    Simple helper script to simplify and bring consistency to all functions
    '''
    vemu.getHeap(emu, 5024*1024)    # initialize heap to 1MB instead of 10k default
    nemu = vemu.NinjaEmulator(emu, fakePEB=True, guiFuncGraphName=funcgraph)
    nemu.addEnvVar(b'INSTALL_DIR', br'C:\foo\bar')
    nemu.addEnvVar(b'OS', br'Windows_NT')

    # create a directory mapping for lots of files/directories
    nemu.addDirectoryMap(br'C:\foo\bar\bin', b'./work/bin/')

    # create a fake file (not on disk) with default attributes
    nemu.addFile(br'C:\foo\bar\bin\blahblah.dll', b'asdflkahsdflkjh')

    nemu.addFilePathMapping(emu.vw.parseExpression('dll_module_in_viv'), "Dll-Module-With-Case_and_hyphes.dll")

    # existing directories 
    nemu.addDirectory(b'C:\\')
    nemu.addDirectory(br'C:\foo')
    nemu.addDirectory(br'C:\foo\bar')
    nemu.addDirectory(br'C:\foo\bar\xml')
    nemu.addDirectory(br'C:\foo\baz')

    # call handlers manually added (a bunch are added by default for common libc/kernel32/etc. funcs)
    nemu.addCallHandler('dll_module_in_viv+0x2a50', LogThing)
        
    nemu.getKernel().registry.setConfigPrimitive(cfgdict)

    stackbase, stacksize = nemu.emu.getMeta('kernel').getStackInfo()
    nemu.emu.addMemoryMap(stackbase + stacksize, 4, 'unknown', b'\0'*4096)

    return nemu

cfgdict = {
        'HKLM': {
            'SYSTEM': {
                'RNG': {
                    "ExternalEntropyCount": 2,
                    #"type:ExternalEntropyCount": nemu.REG_DWORD,
                    "Seed": unhexlify(b'5365656446696c6570760500b62d05517e90cfebc9416c5047c402c6a8fc38c6de55838f8dcf0b244bf83f06aad170e1160c643c8634dc91338fae64b5682440f1b370b5c5ab3e89e6e8ee5b'),
                    "type:Seed": nemu.REG_BINARY,
                    }
                },
            'SOFTWARE': {
                'Microsoft': {
                    'Crypt': {
                        'foo': 4,
                        },
                    'Cryptography': {
                        'RNG': {
                            }
                        }
                    }
                }
            }
        }

def getEmu_<THISFUNCTION>(vw):
    fva = vw.parseExpression('<THISFUNCTION>')
    emu = vw.getEmulator(va=fva, logread=True, logwrite=True, safemem=False)
    nemu = getEmu(emu)

    args = []
    nemu.setupCall(fva, args)

    return nemu
    



import cobra.dcode
import cobra.remoteapp

class SharedVivWorkspace:
    ''' 
    We are basically wrapping a VivCli or VivWorkspace object but adding 
    some Server functionality to support FollowTheLeader
    ''' 
    def __init__(self, vw):
        self.vw = vw 
        
    def __getattr__(self, name):
        '''
        If not defined as SharedVivWorkspace, reach into the VivCli
        '''
        if name not in self.__dict__:
            return getattr(self.vw, name)

    def getLeaderLocations(self, wsname):
        wsinfo = self._req_wsinfo(wsname)
        lock, path, events, users, leaders, leaderloc = wsinfo
        return dict(leaderloc)
        
    def getLeaderSessions(self, wsname):
        wsinfo = self._req_wsinfo(wsname)
        lock, path, events, users, leaders, leaderloc = wsinfo
        return dict(leaders)
        
    #def _fireEvent(self, evt, evtdata):
    #    print("fake _fireEvent(%r, %r)" % (evt, evtdata))
        
        
def shareWorkspace(vw, doref=False):
    # TODO:
    if vw.server:
        logger.warning("Already have a server setup... not yet supported.  vw.server == %r")
        return

    daemon = cobra.CobraDaemon('', 0, msgpack=True)
    daemon.fireThread()
    cobra.dcode.enableDcodeServer(daemon=daemon)
    cobra.remoteapp.shareRemoteApp('vivisect.remote.client', appsrv=vw, daemon=daemon)
    vw.server = SharedVivWorkspace(vw)
    #updateGuiMenus(vw)
        
    print(vw.server)
    return daemon

def updateGuiMenus(vw):
    vwgui = vw.getVivGui()
    for fg in vwgui.getFuncGraphs():
        widg = fg.widget()
        widg.rend_tools.setMenu( widg.getRendToolsMenu() )
            
    for mw in vwgui.getMemoryWidgets():
        widg = mw.widget()
        widg.rend_tools.setMenu( widg.getRendToolsMenu() )




import os
import sys
from setuptools import find_packages, setup

dirn = os.path.abspath(os.path.dirname(__file__))

VERSION = open('VERSION').read().strip()
with open(os.path.join(dirn, 'README.md'), 'r') as fd:
    desc = fd.read()

mods = []
pkgdata = {'scripts': ['scripts/*']}

scripts = []
for s in os.listdir('scripts'):
    if s != '.git':
        scripts.append('scripts/%s'%s)


setup  (name        = 'VivisectION',
        version     = VERSION,
        description = desc,
        long_description=desc,
        long_description_content_type='text/markdown',
        author = 'atlas of d00m',
        author_email = 'atlas@r4780y.com',
        url = 'https://github.com/atlas0fd00m/VivisectION',
        download_url     = 'https://github.com/atlas0fd00m/VivisectION/archive/v%s.tar.gz' % VERSION,
        packages = find_packages(),
        package_data = pkgdata,
        install_requires = [
            'vivisect>=1.0.8'
        ],
        ext_modules = mods,
        scripts = scripts
       )
